#include "lisp.h"

Pointer prin1(Pointer data, Stream stream) {
  switch (TYPE(data)) {
  case CONS:
    fprintf(stream, "(");
    prin1(CAR(data), stream);
    while((data = CDR(data)) != NIL) {
      fprintf(stream, " ");
      prin1(CAR(data), stream);
    }
    fprintf(stream, ")");
    break;
  case NUMBER: fprintf(stream, "%g", NUMBER(data)); break;
  case TABLE: {
    Table* table = &TABLE(data);
    fprintf(stream, "{");
    int num = 0;
    for (unsigned i = 0; i < table->size; i++) {
      TableHash hash = table->data[i].hash;
      for (unsigned j = 0; j < hash.length; j++) {
	TablePair pair = table->data[table->size + hash.offset + j].pair;
	if (num++ > 0) {
	  fprintf(stream, ", ");
	}
	prin1(pair.key, stream);
	fprintf(stream, ": ");
	prin1(pair.value, stream);
      }
    }
    fprintf(stream, "}");
  }
    break;
  case NATIVE_FUNC: fprintf(stream, "NATIVE_FUNC"); break;
  case SPECIAL_FORM: fprintf(stream, "SPECIAL_FORM"); break;
  case FUNC:
    prin1(cons(symbol("fn", sizeof("fn")), FUNC(data).code), stream);
    break;
  case SYMBOL: fprintf(stream, "%.*s", SYMBOL(data).length, SYMBOL(data).data); break;
  case STRING: fprintf(stream, "%.*s", STRING(data).length, STRING(data).data); break;
  case CHAR: fprintf(stream, "\\%1.4s", (char*) &CHAR(data)); break;
  default:
    fprintf(stream,
	    "[MISSING PRINT type: %d, pointer: %d]", 
	    TYPE(data),
	    data);
    break;
  }

  return data;
}

Pointer print(Pointer data, Stream stream) {
  fprintf(stream, "\n");
  prin1(data, stream);
  fprintf(stream, " ");
  fflush(stream);
  return data;
}

Pointer print_fn(Pointer args, Pointer env) {
  Pointer data = CAR(args);
  Pointer streamPtr = CAR(CDR(args));
  Stream stream = streamPtr == NIL
    ? STREAM(environment_get(env, STANDARD_OUTPUT))
    : STREAM(streamPtr);
  return print(data, stream);
}
