#include <math.h>
#include "lisp.h"

Pointer table(size_t size) {
  size = pow(2, ceil(log2(size)));
  Pointer pointer = memory_new(TABLE, sizeof(Table) + size * sizeof(TableData) * 2);
  Table* tbl = &TABLE(pointer);
  tbl->length = 0;
  tbl->size = size;

  for (unsigned i = 0; i < size; i++) {
    tbl->data[i].hash = (TableHash) {0, 0};
  }

  return pointer;
}

Pointer table_get(Pointer pointer, Pointer key) {
  Table* table = &TABLE(pointer);
  if (table->size == 0) return UNDEFINED;
  unsigned i = key & (table->size - 1);
  TableHash hash = table->data[i].hash;

  for (unsigned j = hash.offset + table->size;
       j < hash.offset + table->size + hash.length;
       j++) {
    TablePair pair = table->data[j].pair;
    if (pair.key == key) {
      return pair.value;
    }
  }

  return UNDEFINED;
}

static void addToTable(Table* table, Pointer key, Pointer value) {
  unsigned i = key & table->size - 1;
  TableHash* hash = &table->data[i].hash;
  hash->length++;
  for (unsigned j = table->size + table->length - 1;
       j > table->size + hash->offset;
       j--) {
    table->data[j] = table->data[j - 1];
  }

  for (unsigned j = i + 1; j < table->size; j++) {
    table->data[j].hash.offset++;
  }

  TablePair* pair = &table->data[table->size + hash->offset].pair;
  pair->key = key;
  pair->value = value;
}

Pointer table_set(Pointer pointer, Pointer key, Pointer value) {
  Table* table = &TABLE(pointer);
  unsigned i = key & table->size - 1;
  TableHash hash = table->data[i].hash;

  for (unsigned j = hash.offset + table->size;
       j < hash.offset + table->size + hash.length;
       j++) {
    TablePair* pair = &table->data[j].pair;
    if (pair->key == key) {
      pair->value = value;
      return pointer;
    }
  }

  if (table->length == table->size) {
    table->size = MAX(table->size * 2, 1);
    pointer = memory_resize(pointer, sizeof(Table) + table->size * sizeof(TableData) * 2);
    table = &TABLE(pointer);

    // Move data at the end of the new data section.
    for (unsigned j = table->length; j < table->size; j++) {
      table->data[j + table->size] = table->data[j];
    }

    // Reset the hash section.
    for (unsigned j = 0; j < table->size; j++) {
      table->data[j].hash.offset = 0;
      table->data[j].hash.length = 0;
    }

    // Readd the data in the table.
    for (unsigned j = table->size + table->length; j < table->size * 2; j++) {
      addToTable(table, table->data[j].pair.key, table->data[j].pair.value);
    }
  }

  table->length++;
  addToTable(table, key, value);

  return pointer;
}
